structure LambdaPlusPlus =
struct
  datatype typ =
    TypeVariable of int
  | Arrow of typ * typ
  | Plus of typ * typ
  | Times of typ * typ

  datatype exp =
    Variable of string
  | Lambda of string * exp
  | Apply of exp * exp
  | Tuple of exp * exp
  | First of exp
  | Second of exp
  | Left of exp
  | Right of exp
  | Case of exp * (string * exp) * (string * exp)

  infix 6 ^^
  fun s1 ^^ s2 = s1 ^ " " ^ s2
  local
    val rec tyvars = (fn
      TypeVariable a => [a]
    | (Arrow (t1, t2) | Plus (t1, t2) | Times (t1, t2)) =>
      let
        val vars1 = tyvars t1
        val vars2 = tyvars t2
        val vars2_filtered =
          List.filter
            (fn a => not (List.exists (fn b => a = b) vars1))
            vars2
      in
        vars1 @ vars2_filtered
      end
    )

    fun ith_name i =
      if i < 26
      then Char.toString (chr (i + 97))
      else ith_name (i div 26) ^ ith_name (i mod 26)
  in
    fun pp_typ t =
      let
        val all_tyvars = tyvars t
        val all_tyvars_i =
          List.tabulate (
            (length all_tyvars),
            (fn i => (i, List.nth (all_tyvars, i)))
          )
        val rec pp_typ_help = (fn
            TypeVariable a =>
              ith_name (#1 (valOf (
                List.find (fn (_, b) => b = a) all_tyvars_i
              )))
          | Arrow (t1, t2) => "(" ^ pp_typ_help t1 ^^ "->" ^^ pp_typ_help t2 ^ ")"
          | Plus (t1, t2) => "(" ^ pp_typ_help t1 ^^ "+" ^^ pp_typ_help t2 ^ ")"
          | Times (t1, t2) => "(" ^ pp_typ_help t1 ^^ "*" ^^ pp_typ_help t2 ^ ")"
        )
      in
        pp_typ_help t
      end
  end

  val rec pp_exp = (fn
      Variable x => x
    | Lambda (x, e) => "fn" ^^ x ^^ "=>" ^^ pp_exp e
    | Apply (e1, e2) => "(" ^ pp_exp e1 ^^ pp_exp e2 ^ ")"
    | Tuple (e1, e2) => "(" ^ pp_exp e1 ^ "," ^^ pp_exp e2 ^ ")"
    | First e => "#1" ^^ pp_exp e
    | Second e => "#2" ^^ pp_exp e
    | Left e => "INL" ^^ pp_exp e
    | Right e => "INR" ^^ pp_exp e
    | Case (e, (x1, e1), (x2, e2)) =>
      "case" ^^ pp_exp e ^^ "of" ^^
      "INL" ^^ x1 ^^ "=>" ^^ pp_exp e1 ^^ "|" ^^
      "INR" ^^ x2 ^^ "=>" ^^ pp_exp e2
  )

  structure Parser =
  struct
    val Left_ = Left
    val Right_ = Right
    open ParserCombinators
    open CharParser

    infixr 4 << >>
    infixr 3 &&
    infix  2 -- ##
    infix  2 wth suchthat return guard when
               infixr 1 || <|> ??

    structure LPPDef :> LANGUAGE_DEF =
    struct

      type scanner = char CharParser.charParser

      val commentStart   = SOME "(*"
      val commentEnd     = SOME "*)"
      val commentLine    = NONE
      val nestedComments = true

      val identStart     = CharParser.letter
      val identLetter    = CharParser.alphaNum
      val opStart        = CharParser.oneOf [#"=", #"#", #"|", #",", #"-", #"+", #"*", #":"]
      val opLetter       = CharParser.oneOf [#">", #"1", #"2"]
      val reservedNames  = ["fn", "INL", "INR", "case", "of"]
      val reservedOpNames= ["=>", "|", "#1", "#2", ",", "->", "+", "*"]
      val caseSensitive  = true

    end

    structure LPPP = TokenParser(LPPDef)
    open LPPP

    local
      val c = ref 0
      structure D = StringRedBlackDict
      val table : int D.dict ref = ref D.empty
    in
      fun tvar_rep s =
        case D.find (!table) s of
          SOME n => n
        | NONE => let val n = !c in
                    c := n + 1;
                    table := (D.insert (!table) s n);
                    n
                  end
    end

    val tvar = identifier wth (TypeVariable o tvar_rep)
    fun arrowify [] = raise Fail "you are tearing me apart"
      | arrowify [t] = t
      | arrowify (t1::t2::ts) = Arrow (t1, arrowify (t2::ts))
    fun typ () = separate1 ($typ') (reservedOp "->") wth arrowify
    and typ' () =
          ($typ'' -- (fn t1 => reservedOp "+" >> $typ'' wth (fn t2 => Plus (t1, t2))))
      ||  $typ''
    and typ'' () =
          ($typ''' -- (fn t1 => reservedOp "*" >> $typ''' wth (fn t2 => Times (t1, t2))))
      ||  $typ'''
    and typ''' () = tvar <|> parens ($typ)

    val full_type = $typ << (eos ?? "end of stream")

    val var = identifier wth Variable
    fun exp () = chainl1 ($exp') (succeed Apply)
    and exp' () =
      (parens ($exp) || $tuple)
      <|> var
      <|> $lambda
      <|> $first
      <|> $second
      <|> $left
      <|> $right
      <|> $case_analysis
    and lambda () =
      reserved "fn" >>
      identifier -- (fn x =>
        reservedOp "=>" >>
        $exp wth (fn e =>
          Lambda (x, e)
        )
      )
    and tuple () = parens (
        $exp -- (fn e1 =>
          reservedOp "," >>
          $exp wth (fn e2 =>
            Tuple (e1, e2)
          )
        )
      )
    and first () = reservedOp "#1" >> $exp wth First
    and second () = reservedOp "#2" >> $exp wth Second
    and left () = reserved "INL" >> $exp wth Left_
    and right () = reserved "INR" >> $exp wth Right_
    and case_analysis () =
      reserved "case" >>
      $exp -- (fn e =>
        reserved "of" >>
        reserved "INL" >>
        identifier -- (fn x1 =>
          reservedOp "=>" >>
          $exp -- (fn e1 =>
            reservedOp "|" >>
            reserved "INR" >>
            identifier -- (fn x2 =>
              reservedOp "=>" >>
              $exp wth (fn e2 =>
                Case (e, (x1, e1), (x2, e2))
              )
            )
          )
        )
      )
    val program = $exp << (eos ?? "end of stream")
  end

  val parse_exp = CharParser.parseString Parser.program
  val parse_typ = CharParser.parseString Parser.full_type
end
